//  ----------------------------------------------------------
//
//  Copyright (C) 2002 Brad Wasson <bard@systemtoolbox.com>
//
//  This file is part of 3ddesktop.
//
//  3ddesktop is free software; you can redistribute it and/or modify it
//  under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2, or (at your option)
//  any later version.
//
//  3ddesktop is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with 3ddesktop; see the file COPYING.   If not, write to
//  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <string.h>
#include <getopt.h>

#include "3ddesk.h"

static void 
usage_and_bye (void)
{
    printf ("usage: 3ddesk [ OPTIONS ]\n\n"
            "Activates the 3ddesktop.  Where\n"
            "     --nozoom    Disable the zoom out\n"
            "     --gotoprev  Goto the previous desktop\n"
            "     --gotonext  Goto the next desktop\n"
            "     --goto=x    Goto desktop 'x' (1 thru max)\n\n"
        );
    exit (1);
}

int verbose = 0;

int main(int argc, char **argv)
{
    struct threedeedesk_command msg;
    msg.mtype = MTYPE_ACTIVATE;
    msg.goto_face = NO_GOTO;
    msg.zoom = 1;
    msg.face_type = 0;

    static struct option long_options[] = {
        {"nozoom", 0, 0, 0},
        {"goto", 1, 0, 0},
        {"gotonext", 0, 0, 0},
        {"gotoprev", 0, 0, 0},
        {"mode", 1, 0, 0},
        {0, 0, 0, 0}
    };

    char c;
    int option_index;
    while (1) {

        c = getopt_long (argc, argv, "v",
                         long_options, &option_index);
        if (c == -1)
            break;

        switch (c) {
        case 0: // long option

            switch (option_index) {
            case 0: // 'nozoom'
                msg.zoom = 0;
                break;
            case 1: // 'goto'
                msg.goto_face = atoi (optarg);
                if (msg.goto_face < 1) 
                    usage_and_bye();
                break;
            case 2: // 'gotonext'
                msg.goto_face = GOTO_NEXT_FACE;
                break;
            case 3: // 'gotoprev'
                msg.goto_face = GOTO_PREV_FACE;
                break;
            case 4: // 'mode'
                if (strcmp(optarg, "cylinder") == 0) {
                    msg.face_type = FACE_TYPE_CYL;
                } else if (strcmp(optarg, "linear") == 0) {
                    msg.face_type = FACE_TYPE_LINEAR;
                } else if (strcmp(optarg, "viewmaster") == 0) {
                    msg.face_type = FACE_TYPE_VIEWMASTER;
                } else if (strcmp(optarg, "carousel") == 0) {
                    msg.face_type = FACE_TYPE_CAROUSEL;
                } else if (strcmp(optarg, "priceisright") == 0) {
                    msg.face_type = FACE_TYPE_PRICEISRIGHT;
                } else {
                    usage_and_bye();
                }
                break;
            }
            break;
        case 'v':
            verbose = 1;
            break;
        default:
            usage_and_bye();
            break;
        }
    }

    int msqid;
    key_t key;
    int rc;

    key = ftok("/bin/sh", 222);
    if (key == (key_t)-1) {
        fprintf(stderr,"ftok failed: %d: %s\n", errno, strerror(errno));
        exit (1);
    }
    
    msqid = msgget(key, 0644);
    if (msqid < 0) {
        fprintf(stderr, "Can't get message queue: %s\n", strerror(errno));
        fprintf(stderr, "3ddeskd daemon started?\n");
        exit (1);
    }
    
    rc = msgsnd(msqid, (struct msgbuf *)&msg, sizeof(msg) - sizeof(long), 0);
    if (rc < 0) {
        fprintf(stderr, "Failed to send daemon message\n");
        exit (1);
    }

    return 0;
}

