//  ----------------------------------------------------------
//  3d Desktop
//
//  Allows you to switch virtual desktops (viewports) in several
//  3-dimensional ways.
// 
//  I started with a program called "OpenGL cube demo" by Chris
//  Halsall.  It has changed quite dramatically since then but it was
//  a *huge* help to me.  This is my first OpenGL program.  It was
//  coded up starting around Feb/March 2002.  Please provide feedback,
//  fixes, improvements etc. to bard@systemtoolbox.com
//
//  ----------------------------------------------------------
//
//  Copyright (C) 2002 Brad Wasson <bard@systemtoolbox.com>
//
//  This file is part of 3ddesktop.
//
//  3ddesktop is free software; you can redistribute it and/or modify it
//  under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2, or (at your option)
//  any later version.
//
//  3ddesktop is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with 3ddesktop; see the file COPYING.   If not, write to
//  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include <signal.h>

#define _GNU_SOURCE
#include <unistd.h>
#include <dirent.h>  // for PATH_MAX?
#include <getopt.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <syslog.h>

// STL
using namespace std;
#include <iostream>
#include <list>

#include <GL/glx.h>
#include <GL/gl.h>   // OpenGL itself.
#include <GL/glu.h>  // GLU support library.
#include <GL/glut.h> // GLUT support library.

#include <X11/extensions/xf86vmode.h>
#include <X11/keysym.h>


#include "3ddesk.h"

#include "move.hpp"
#include "face.hpp"
#include "faceset.hpp"
#include "xutil.hpp"

#include "util.h"

#include "win.hpp"

typedef unsigned char uchar;

// Some global variables.... 
int do_screenshot = 1;
int do_vpswitch = 1;
int do_fullscreen = 1;

int use_tmp = 1;
int use_context_switch = 1;
int use_breathing = 1;
int daemonized = 0;

float default_z_offset_distance = 0;


VDesktops vdesks;

// time is seconds before random movements start ;)
int random_fun_delay = 15;

int verbose = 0;
int show_fps = 0;
int entry_exit_movement = 1;
int clear_other_screens = 0;

int msqid = -1;

int goto_face = NO_GOTO;

// Our display mode settings.
int Light_On = 1;
int Blend_On = 0;
int Texture_On = 1;
int Alpha_Add = 0;

int Curr_TexMode = 1;
char *TexModesStr[] = {"GL_DECAL","GL_MODULATE","GL_BLEND","GL_REPLACE"};
GLint TexModes[] = {GL_DECAL,GL_MODULATE,GL_BLEND,GL_REPLACE};


GLint texture[MAX_FACES];
GLint digit_text[MAX_FACES];

unsigned int texture_size = 512;  // in bytes

// delay between updates of the movements (in milliseconds)
float animation_step = 10;
long last_time;  // for movement timing


// time of last cbrenderscene
long last_renderscene_time = 0;
// time of last keypress 
float last_keypress = 0;
// time of next "ghost keypress" when having fun
float next_action = 0;


//struct gnome_area_info area_info;
//struct gnome_workspace_info workspace_info;


// set default
int face_type = FACE_TYPE_CAROUSEL;


int initiate_destroy = 0;   // starts the z offset zoom in

FaceSet *faces;

#if defined(__FreeBSD__)
char working_dir[MAXNAMLEN + 1];
char __base_dir[MAXNAMLEN + 1] = "/usr/share/3ddesktop";
char pidfile[MAXNAMLEN + 1] = "";
#else
char working_dir[PATH_MAX];
char __base_dir[PATH_MAX] = "/usr/share/3ddesktop";
char pidfile[PATH_MAX] = "";
#endif

char *base_dir = __base_dir;




// lights

// diffuse light
//float light_ambient[]=  { 1.0f, 1.0f, 1.0f, 1.0f };
float light_diffuse[]=  { 1.0f, 1.0f, 1.0f, 1.0f }; 
float light_position[]= { 0.0f, 0.0f, 17.0f, 1.0f };


Movement breathing;  // changes the ambient light to look like breathing
float ambient_value = 2.0;


// ambient light
float light_ambient2[]=  { 1.0f, 1.0f, 1.0f, 1.0f };
//float light_diffuse2[]=  { 1.0f, 1.0f, 1.0f, 1.0f }; 
float light_position2[]= { 0.0f, 0.0f, 17.0f, 1.0f };


GLWindow GLWin;
Bool done;


enum state_t {
    STATE_ENTRY,
    STATE_EXIT,
    STATE_NORMAL,
    STATE_GOTO
};

state_t state = STATE_ENTRY;

// prototypes 
static void gl_init (void);
static void load_digit_textures();
static void setup_faces (float z_offset_distance, int current_face, int face_count);
static void setup_goto_face(int goto_face, int current_face, int face_count);
static int main_loop(void);
static int make_pidfile(void);
static int daemon_init (void);

static void calc_fps(void);

static void render_scene(void);
static void idle_func(void);
static void mouse_pressed(unsigned int button);
static void key_pressed(KeySym key);


// this is a hack... I tried to do it OO, but it doesn't seem elegant
// or clean... I was hoping to encapsulate the process of going to
// another face automatically (no messy globals etc).  I'll consider
// this a first wack at the problem...
class Task {
private:
    long gotime;
public:
    Task (long gt) {
        gotime = gt;
    }
    virtual ~Task () { };
    int check_and_run(void) {
        if ( gotime <= get_milli_time() ) {
            function();
            return 1;
        }
        return 0;
    }

    virtual void *function(void) = 0;
}; 

list<Task *> tasks;


// This starts the final zoom out and is schedule by "GotoFaceTask" to
// run after a face has been jumped to.
class GotoEndTask : public Task {
    int zoom;
    FaceSet *faces;
public:
    GotoEndTask (long gt, int z, FaceSet *fs) : Task (gt) {
        zoom = z;
        faces = fs;
    }
    void *function (void) {
        if (zoom)
            faces->exit_movement_start();
        initiate_destroy = 1;
        return NULL;
    }
};


// This is scheduled to run when the object is to automatically go to
// another face.  It is scheduled in setup_goto_face() to start the
// goto face movement after the entry/exit (zoom out) movement is
// complete.  In addition it schedules the exit movement (zoom in)
// when the goto face operation is complete.
class GotoFaceTask : public Task {
    FaceSet *faces;
    int face_to_goto;
    int zoom;
public:
    GotoFaceTask (int face, long gt, int z, FaceSet *fs) : Task(gt) {
        face_to_goto = face;
        zoom = z;
        faces = fs;
    }

    void *function(void) {
        float speed = faces->goto_face(face_to_goto);
        
        long t = get_milli_time() 
                 + (long)(speed * animation_step)
                 + 400;  // not an exact science ;)
        
        GotoEndTask *end_task = new GotoEndTask (t,zoom,faces);
        if (!end_task) {
            msgout (ERROR, "out of memory\n");
            end_program(-1);
        }
        
        tasks.push_back((Task *)end_task);
        return NULL;
    }
};


GotoFaceTask *goto_face_task;




static void 
usage_and_bye (void)
{
    msgout(INFO, 
            "usage: 3ddeskd [ --mode={carousel   | cylinder | linear |\n"
            "                         viewmaster | priceisright} ]\n"
            "              [ --dir=/path     ] \n"
            "              [ --workspaces    ] \n"
            "              [ --depth=#       ] \n"
            "              [ --randdelay=num ] \n"
            "              [ --fps           ] \n"
            "              [ --texturesize=# ] \n"
            "              [ --clearshm      ] \n"
            "              [ --delshm        ] \n"
            "              [ --fastest       ] \n"
            "              [ -v ]\n\n"
            "  where:\n"
            "    --mode        is the way the desktops are shown\n"
            "    --dir         the install dir if different than /usr/share/3ddesktop\n"
            "    --workspaces  switch between workspace desktops instead of grid areas\n"
            "    --depth       specify how much to zoom out\n"
            "    --randdelay   seconds before random movement (0 for none) 15 default\n"
            "    --fps         shows the frames per second\n"
            "    --texturesize size of textures (larger for more detail) in powers of 2\n" 
            "    --fastest     hog the CPU for smoothest operation\n"
            "    -v            verbose/debug information\n\n");
    exit (1);
} // END usage_and_bye


// ------
// The main() function.  Inits OpenGL.  Calls our own init function,
// then passes control onto OpenGL.
int main(int argc,char **argv)
{
    char c;
    int option_index;
    int tmpint;
    
    // this indicates we should use the "layered" workspaces instead of
    // horizontal/vertical areas
    int use_workspaces = 0;
    // use system("sawfish-client ...") for sawfish without gnome
    int sawfish_only = 0;
 
    int acquire = 0;

    static struct option long_options[] = {
        {"mode", 1, 0, 0},
        {"fps", 0, 0, 0},
        {"dir", 1, 0, 0},
        {"randdelay", 1, 0, 0},
        {"workspaces", 0, 0, 0},
        {"texturesize", 1, 0, 0},
        {"fastest", 0, 0, 0},
        {"depth", 1, 0, 0},
        {"win", 0, 0, 0},
        {"sawfish", 0, 0, 0},
        {"acquire", 0, 0, 0},
        {0, 0, 0, 0}
    };

    float tmpfloat;

    while (1) {

        c = getopt_long (argc, argv, "v",
                         long_options, &option_index);
        if (c == -1)
            break;

        switch (c) {
        case 0: // 'mode'

            //msgout (DEBUG, "option %s", long_options[option_index].name);
            switch (option_index) {
            case 0:
                if (!optarg)
                    usage_and_bye();

                if (strcmp(optarg, "cylinder") == 0) {
                    face_type = FACE_TYPE_CYL;
                } else if (strcmp(optarg, "linear") == 0) {
                    face_type = FACE_TYPE_LINEAR;
                } else if (strcmp(optarg, "viewmaster") == 0) {
                    face_type = FACE_TYPE_VIEWMASTER;
                } else if (strcmp(optarg, "carousel") == 0) {
                    face_type = FACE_TYPE_CAROUSEL;
                } else if (strcmp(optarg, "priceisright") == 0) {
                    face_type = FACE_TYPE_PRICEISRIGHT;
                } else {
                    usage_and_bye();
                }
                break;

            case 1: // 'fps'
                show_fps = 1;
                break;
            case 2: // 'dir'
                strncpy(base_dir, optarg, sizeof(__base_dir));
                break;
            case 3: // 'randdelay'
                tmpint = atoi (optarg);
                if (0 <= tmpint && tmpint < 60)
                    random_fun_delay = tmpint;
                break;
            case 4: // 'workspaces'
                use_workspaces = 1;
                break;
            case 5: // 'texturesize'
                texture_size = atoi (optarg);
                
                switch (texture_size) {
                case  16: case  32: case   64: case  128:
                case 256: case 512: case 1024: case 2048:
                    break;
                default:
                    // I now use gluBuild2DMipmaps so this isn't
                    // really necessary because it converts it however
                    // it is sooooo slow.  To use glTexImage2D it is
                    // necessary.
                    msgout (ERROR, "Texture size must be a power of 2 (such as 256, 512, 1024 ...)\n");
                    end_program(-1);
                    break;
                }
                break;
            case 6: // 'fastest'
                use_context_switch = 0;
                break;
            case 7: // 'depth'
                tmpfloat = (float)atof(optarg);
                if (3.5 <= tmpfloat && tmpfloat < 70.0)
                    default_z_offset_distance = tmpfloat;
                break;
            case 8: // 'win'
                do_fullscreen = 0;
                break;
            case 9: // 'sawfish'
                sawfish_only = 1;
                break;
            case 10: // 'acquire'
                acquire = 1;
                break;
            }
            break;
        case 'v':
            verbose = 1;
            msgout (INFO, "Verbose is ON\n");
            break;
        default:
            usage_and_bye();
            break;
        }
    }  // end while args



    // creates ~/.3ddesktop or /tmp/.3ddesktop-user if not alread there
    // used for pid file but no longer the screenshot
    create_working_dir_if_necessary ();

    // there is a race between here and marking that I am running
    // but it should be small.  We must separate these because the
    // semaphore ownership is not inherited if we daemonize.
    if (check_if_i_am_running()) {
        msgout (ERROR, "3ddesktop is already running.  bye\n");
        exit (1);
    }
        
    vdesks.init (use_workspaces, sawfish_only);

    int face_count = 1;
    int current_face = 0;

    vdesks.get_vdesktop_info (&current_face, &face_count);


    if (face_count > MAX_FACES) {
        msgout (ERROR, "Can't support this many virtual desktops (%d), MAX is %d.\n", 
                face_count, MAX_FACES);
        end_program (1);
    }

    if (face_count <= 1) {
        msgout (ERROR, "You only have %d virtual desktop(s)!  Need at least 2 to switch between them...\n", face_count);
        end_program (1);
    }

    // the actual acquire is done below but the message must be here,
    // before we daemonize
    if (acquire)
        msgout(INFO, 
               "\n"
               "================================================================\n"
               "3ddesktop will be acquiring images in one moment. Please wait...\n"
               "================================================================\n\n");


    // become a daemon 
    if (!verbose) {
        openlog ("3ddeskd", LOG_CONS, LOG_DAEMON);
        
        if (daemon_init() < 0) {
            msgout(ERROR, "Failed to daemonize.\n");
            end_program(-1);
        }

        daemonized = 1;
    }

    struct sigaction sa;
    sigset_t mask;

    sigemptyset (&mask);
    memset (&sa, 0, sizeof (struct sigaction));

    sa.sa_mask = mask;
    sa.sa_handler = end_program;

    sigaction (SIGTERM, &sa, NULL);
    sigaction (SIGINT, &sa, NULL);
    sigaction (SIGQUIT, &sa, NULL);


    // create the message queue we will listen on
    key_t key;
    key = ftok("/bin/sh", 222);
    if (key == (key_t)-1) {
        msgout(ERROR,"ftok failed: %d: %s\n", errno, strerror(errno));
        end_program(-1);
    }
    
    msqid = msgget(key, 0644 | IPC_CREAT);
    if (msqid == -1) {
        msgout(ERROR, "Could not create message queue! %s\n", strerror(errno));
        end_program(-1);
    }
        

    if (flag_that_i_am_running()) {
        // (if we've daemonized this goes to syslog which is not easily seen)
        msgout(ERROR, "Oops, someone beat us to start up!!  Amazing!\n");
        end_program(-1);
    }

    if (make_pidfile() < 0)
        end_program(-1);


    // initialize opengl

    // our own init function.
    gl_init();

    GLWin.open_display(verbose);


    GLWin.create_GL_window("Enhanced 3-dimensional workspace visualization magic machine", 
                           do_fullscreen);

    // Loads up the correct perspective matrix
    GLWin.resize_scene();

    GLWin.hide_GL_window();

    for (int i = 0; i < face_count; i++) 
        texture[i] = -1;

    load_digit_textures();

    setup_faces(default_z_offset_distance, current_face, face_count);


    // loop through desktops and capture their images...
    if (acquire) {

        sleep(1);

        int i;
        for (i = 0; i < face_count; i++) {

            vdesks.set_vdesktop (i);
            
            if (GLWin.grab_screenshot_data() < 0) {
                msgout (ERROR, "couldn't grab screen image\n");
                end_program(-1);
            }
            GLWin.load_screenshot_data((GLuint *)&texture[i]);
        }

        // go back to current
        vdesks.set_vdesktop (current_face);

    }

    if (entry_exit_movement)
        faces->entry_movement_start();

    return main_loop();

}  // END main

static int daemon_init (void)
{
    pid_t pid;

    msgout (DEBUG, "daemon init\n");
    
    if ( (pid = fork()) < 0) {
        msgout (ERROR, "daemon_init: fork failed: %s\n", strerror(errno));
        return -1;
    } else if (pid != 0) {
        msgout (INFO, "Daemon started.  Run 3ddesk to activate.\n");
        exit(0);  /* bye parent */
    }

    /* child continues */

    setsid();
    
    chdir ("/");

    umask (0);

    return(0);
} /* END daemon_init */


static int make_pidfile (void) 
{
    pid_t pid;

    sprintf (pidfile, "%s/pid", working_dir);

    FILE *fh = fopen (pidfile, "w");
    if (fh == NULL) {
        msgout (ERROR, "Could not open %s: %s\n", 
                pidfile, strerror(errno));
        return -1;
    }

    pid = getpid();
    
    fprintf (fh, "%d", pid);
    
    fclose (fh);
    
    return 0;
} /* END make_pidfile */


static void 
setup_goto_face(int goto_face, int current_face, int face_count)
{
    long t;
    int face = goto_face - 1;   // actual faces are 0 thru face_count - 1

    if (goto_face == GOTO_NEXT_FACE) {
        face = current_face + 1;
        if (face >= face_count) face = 0;

    } else if (goto_face == GOTO_PREV_FACE) {
        face = current_face - 1;
        if (face < 0) face = face_count - 1;
    }

    
    msgout (DEBUG,"ee move is %d\n", entry_exit_movement);
    // time it takes to change faces - longer for entry/exit
    // movement - shorter when not
    if (entry_exit_movement)
        t = get_milli_time() + 600;
    else
        t = get_milli_time() + 100;

    goto_face_task = new GotoFaceTask (face,
                                       t,
                                       entry_exit_movement,
                                       faces); 
        
    if (!goto_face_task) {
        msgout (ERROR, "out of memory\n");
        end_program(-1);
    }

    tasks.push_back((Task *)goto_face_task);

} // END setup_goto_faces

static int
wait_for_activation(void)
{
    struct threedeedesk_command msg;
    int rc;
    int done = 0;
    int face_count = 1;
    int current_face = 0;

    goto_face = NO_GOTO;
    do {
        // should block until message received
        rc = msgrcv(msqid, (struct msgbuf *)&msg, sizeof(msg)- sizeof(long), 0, 0);
        if (rc < 0) {
            msgout (ERROR, "msgrcv failed: %s\n", strerror(errno));
            end_program(-1);
        }

        msgout (DEBUG, "MSG: %ld\n", msg.mtype);


        // must get this now in case the current desktop was change by
        // other means
        vdesks.get_vdesktop_info (&current_face, &face_count);

        msgout(DEBUG, "####### current %d\n", current_face);
        faces->set_current_face (current_face);

        switch (msg.mtype) {
        case MTYPE_ACTIVATE:
            entry_exit_movement = msg.zoom;
            goto_face = msg.goto_face;

            // check bounds on goto face
            if (goto_face > NO_GOTO) {
                if (goto_face == current_face + 1)
                    continue;

                if ( goto_face < 1
                     || goto_face > face_count )  
                {
                    msgout (DEBUG, "Ignoring out of bounds goto face of %d\n", goto_face);
                    continue;
                }
            }

            // we must destroy our current face and create another one
            // if the indicated face_type is different or the number
            // of faces has changed
            if ( msg.face_type 
                 && 
                 (msg.face_type != face_type) )
            {
                msgout(DEBUG, "destroy faceset for %d\n", face_type);
                face_type = msg.face_type;
                delete faces;
                setup_faces (default_z_offset_distance, 
                             current_face, face_count);

            } else if ( face_count != faces->get_number_of_faces() ) {
                msgout (DEBUG, "++++ face_count changed from %d to %d\n", 
                        faces->get_number_of_faces(), face_count);

                delete faces;
                setup_faces (default_z_offset_distance, 
                             current_face, face_count);
            }

            done = 1;
            break;
        default:
            msgout (ERROR, "Unknown message! mytpe = %d\n", msg.mtype);
            break;
        }
    } while (!done);


    msgout (DEBUG, "Activated\n");
    
    return 0;
}


static int 
main_loop(void)
{
    XEvent event;
    KeySym key;
    Bool quit = False;
    int current_face;

    while(!quit) {

        if ( wait_for_activation() )
            break;
        
        current_face = faces->get_current_face();

        if (do_screenshot) {
        
            if (GLWin.grab_screenshot_data() < 0) {
                msgout (ERROR, "couldn't grab screen image\n");
                end_program(-1);
            }
            GLWin.load_screenshot_data((GLuint *)&texture[current_face]);

            faces->make_display_list();
        }

        GLWin.unhide_GL_window();

        initiate_destroy = 0;

        if (entry_exit_movement)
            faces->entry_movement_start();

        if (goto_face) 
            setup_goto_face(goto_face, faces->get_current_face(),
                            faces->get_number_of_faces() );

        done = False;

        // finally this is so the "fun" behavior doesn't start right away
        last_keypress = get_milli_time();
    
        /* wait for events*/ 
        while (!done)
        {
            /* handle the events in the queue */
            while (XPending(GLWin.get_display()) > 0)
            {
                XNextEvent(GLWin.get_display(), &event);
                switch (event.type)
                {
                case Expose:
                    if (event.xexpose.count != 0)
                        break;
                    render_scene();
                    break;
                case ConfigureNotify:
                    /* call resizeGLScene only if our window-size changed */
                    if (((unsigned int)event.xconfigure.width != GLWin.get_width()) || 
                        ((unsigned int)event.xconfigure.height != GLWin.get_height()))
                    {
                        GLWin.resize_scene(event.xconfigure.width,
                                           event.xconfigure.height);
                    }
                    break;
                    /* exit in case of a mouse button press */
                case ButtonPress:
                    mouse_pressed (event.xbutton.button);
                    break;
                case KeyPress:
                    key = XLookupKeysym(&event.xkey, 0);
                    key_pressed (key);
                    break;
                case ClientMessage:
                    if (*XGetAtomName(GLWin.get_display(), event.xclient.message_type) == 
                        *"WM_PROTOCOLS")
                    {
                        msgout(INFO, "Exiting sanely...\n");
                        done = True;
                        quit = True;
                    }
                    break;
                default:
                    break;
                }
            }

            if (use_context_switch)
                idle_func();
            else
                render_scene();

        }

        GLWin.hide_GL_window();
        //GLWin.kill_GL_window();

    }


    return 0;

}  // END main_loop

void end_program (int rv)
{
    flag_that_i_am_NOT_running();

    if (pidfile[0])
        unlink(pidfile);

    if (msqid != -1)
        if (msgctl(msqid, IPC_RMID, NULL) == -1) {
            msgout(ERROR, "Could not delete message queue: %s\n", strerror(errno));
        }
    
    exit (rv);
}


// ------
// Frames per second (FPS) statistic variables and routine.
#define FRAME_RATE_SAMPLES 50
int   FrameCount = 0;
float FrameRate = 0;

static void 
calc_fps(void)
{
    static clock_t last=0;
    clock_t now;
    float delta;

    if (++FrameCount >= FRAME_RATE_SAMPLES) {
        now  = clock();
        delta= (now - last) / (float) CLOCKS_PER_SEC;
        last = now;

        FrameRate = FRAME_RATE_SAMPLES / delta;
        FrameCount = 0;
    }
} // END calc_fps



// ------
// Routine which actually does the drawing
static void
render_scene(void)
{
    //char buf[80]; // For our strings.
    long current_time;


    // Enables, disables or otherwise adjusts as 
    // appropriate for our current settings.

    if (Texture_On)
        glEnable(GL_TEXTURE_2D);
    else
        glDisable(GL_TEXTURE_2D);

    if (Light_On) 
        glEnable(GL_LIGHTING);
    else 
        glDisable(GL_LIGHTING);

    if (Alpha_Add)
        glBlendFunc(GL_SRC_ALPHA,GL_ONE); 
    else
        glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);

    // If we're blending, we don't want z-buffering.
    if (Blend_On)
        glDisable(GL_DEPTH_TEST); 
    else
        glEnable(GL_DEPTH_TEST); 


    // Need to manipulate the ModelView matrix to move our model around.
    glMatrixMode(GL_MODELVIEW);

    // Reset to 0,0,0; no rotation, no scaling.
    glLoadIdentity(); 


    current_time = get_milli_time();

    // this gives us semi-precise/consistent timing of the movements
    // 
    if (current_time - last_time > animation_step) {

        if (initiate_destroy) {
            if (!entry_exit_movement || faces->is_exit_movement_done()) {
                // quit (the z offset was changed to match the screen)
                int last_area = faces->get_current_face();
                
                msgout (DEBUG, "You left on area %d\n", last_area);

                if (do_vpswitch)
                    vdesks.set_vdesktop (last_area);
                

                done = True;
                return;
                
            } else {
                if (entry_exit_movement)
                    faces->entry_exit_move();
            }
               
        } else {

            // we are not in the zoom in but we might be in the zoom
            // out so if we are zooming check it
            if (entry_exit_movement && !faces->is_entry_movement_done()) {
                faces->entry_exit_move();
            }
            
        }

        faces->do_movement();

        last_time = get_milli_time();

    } // end animation step
 

    // FIXME: use "Task" to accomplish this?

    // after some seconds start some fun ;)
    if (random_fun_delay &&
        current_time - last_keypress > random_fun_delay * 1000) 
    {

        //msgout(DEBUG, "having some fun...\n");
        if (current_time >= next_action) {
            //if ( rand() < (RAND_MAX/2))
            //    faces->goto_previous();
            //else 
            //    faces->goto_next();

            faces->goto_random_face();
            
            next_action = current_time + ( get_random(350, 600)  );
        }
    }


    // check "tasks"
    list<Task *>::iterator k;
    for (k = tasks.begin(); k != tasks.end(); ++k)
        if ((*k)->check_and_run())
            tasks.erase(k++);



    // adjust the faces the necessary amount
    faces->do_position_GL();


    // Clear the color and depth buffers.
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);


    float ones[] = { 1.0, 1.0, 1.0, 1.0 };

    glMaterialfv(GL_FRONT, GL_DIFFUSE, ones);
    glMaterialfv(GL_FRONT, GL_SPECULAR, ones);
    glMaterialf(GL_FRONT, GL_SHININESS, 100.0);

    faces->render();

    glLoadIdentity();

#if 0
    float spot_direction[] = { 0.0, 0.0, -1.0 };
    glLightfv(GL_LIGHT1, GL_SPOT_DIRECTION, spot_direction);
    glLighti(GL_LIGHT1, GL_SPOT_EXPONENT, 30);
    glLighti(GL_LIGHT1, GL_SPOT_CUTOFF, 180);
#endif

    // lighting (light1 is diffuse only)
    glLightfv(GL_LIGHT1, GL_POSITION, light_position);
    //glLightfv(GL_LIGHT1, GL_AMBIENT,  Light_Ambient);
    glLightfv(GL_LIGHT1, GL_DIFFUSE,  light_diffuse);
    glLightfv(GL_LIGHT1, GL_SPECULAR,  ones);
    glEnable (GL_LIGHT1);

    if (use_breathing) {
        if (breathing.at_destination()) {
            breathing.init(&ambient_value, 1.7, 3.0, 200, Movement::MOVE_TYPE_SINE);
        } else {
            breathing.change_a_bit();
            light_ambient2[0] = light_ambient2[1] = light_ambient2[2] = ambient_value;
        }
    }
        

    // lighting (light2 is for ambient only)
    glLightfv(GL_LIGHT2, GL_POSITION, light_position2);
    glLightfv(GL_LIGHT2, GL_AMBIENT,  light_ambient2);
    //glLightfv(GL_LIGHT2, GL_DIFFUSE,  Light_Diffuse2); 
    glEnable (GL_LIGHT2);

#if 0

    // We need to change the projection matrix for the text rendering.  
    glMatrixMode(GL_PROJECTION);

    // But we like our current view too; so we save it here.
    glPushMatrix();

    // Now we set up a new projection for the text.
    glLoadIdentity();
    glOrtho(0,GLWin.get_width(),0,GLWin.get_height(),-1.0,1.0);

    // Lit or textured text looks awful.
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_LIGHTING);

    // We don't want depth-testing either.
    glDisable(GL_DEPTH_TEST); 


    // Now we want to render the calulated FPS at the top.
   
    // To ease, simply translate up.  Note we're working in screen
    // pixels in this projection.
   
    if (show_fps) {
        glTranslatef(6.0f, GLWin.get_height() - 14,0.0f);
        
        // Make sure we can read the FPS section by first placing a 
        // dark, mostly opaque backdrop rectangle.
        glColor4f(0.2,0.2,0.2,0.75);
        
        glBegin(GL_QUADS);
        glVertex3f(  0.0f, -2.0f, 0.0f);
        glVertex3f(  0.0f, 12.0f, 0.0f);
        glVertex3f(140.0f, 12.0f, 0.0f);
        glVertex3f(140.0f, -2.0f, 0.0f);
        glEnd();
        
        glColor4f(0.9,0.2,0.2,.75);
        sprintf(buf,"FPS: %f F: %2d", FrameRate, FrameCount);
        glRasterPos2i(6,0);
        ourPrintString(GLUT_BITMAP_HELVETICA_12,buf);
    }

    // Done with this special projection matrix.  Throw it away.
    glPopMatrix();
    
#endif

    // All done drawing.  Let's show it.
    glXSwapBuffers(GLWin.get_display(), GLWin.get_window());

    // And collect our statistics.
    if (show_fps)
        calc_fps();

    last_renderscene_time = get_milli_time();

} // END render_scene



static void
mouse_pressed(unsigned int button)
{
    last_keypress = get_milli_time();

    //msgout (DEBUG, "button pressed %d - 0x%x\n", button, button);

    switch (button) {
    case 1: // left
    case 5: // mouse wheel down
        faces->goto_previous();
        break;

    case 3: // right
    case 4: // mouse wheel up
        faces->goto_next();
        break;

    case 2: // middle
        if (entry_exit_movement)
            faces->exit_movement_start();

        initiate_destroy = 1;
        // the z_offset translation starts and is carried out by the
        // render function where we exit when it completes
        break;
    }
} // END mouse_pressed


// ------
// Callback function called when a normal key is pressed.

static void
key_pressed(KeySym key)
{
    last_keypress = get_milli_time();

    next_action = 0;

    switch (key) {
    case 130: case 98: // B - Blending.
        Blend_On = Blend_On ? 0 : 1; 
        if (!Blend_On)
            glDisable(GL_BLEND); 
        else
            glEnable(GL_BLEND);
        break;

    case 108: case 76:  // L - Lighting
        Light_On = Light_On ? 0 : 1; 
        break;

    case 109: case 77:  // M - Mode of Blending
        if ( ++ Curr_TexMode > 3 )
            Curr_TexMode=0;
        glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,TexModes[Curr_TexMode]);
        break;

    case 116: case 84: // T - Texturing.
        Texture_On = Texture_On ? 0 : 1; 
        break;

    case 97: case 65:  // A - Alpha-blending hack.
        Alpha_Add = Alpha_Add ? 0 : 1; 
        break;

    case 114: case 82: // R - random
        faces->goto_random_face();
        break;

#if 1 // for debugging best light position...
    case 119: // w
        light_position2[0] -= 0.5;
        msgout (DEBUG, "light x at %f\n", light_position2[0]);
        break;
    case 101: // e
        light_position2[0] += 0.5;
        msgout (DEBUG, "light x at %f\n", light_position2[0]);
        break;
    case 115: // s 
        light_position2[1] -= 0.5;
        msgout (DEBUG, "light y at %f\n", light_position2[1]);
        break;
    case 100: // d 
        light_position2[1] += 0.5;
        msgout (DEBUG, "light y at %f\n", light_position2[1]);
        break;
    case 120: // x
        light_position2[2] -= 0.5;
        msgout (DEBUG, "light z at %f\n", light_position2[2]);
        break;
    case 99: // c
        light_position2[2] += 0.5;
        msgout (DEBUG, "light z at %f\n", light_position2[2]);
        break;
#endif

    case 48:
        faces->goto_face(9);
        break;
        
    case 49:
    case 50:
    case 51:
    case 52:
    case 53:
    case 54:
    case 55:
    case 56:
    case 57:
        faces->goto_face(key - 49);
        break;

#define  KEY_ENTER  0xff0d
#define  KEY_UP     0xff52
#define  KEY_DOWN   0xff54
#define  KEY_RIGHT  0xff53
#define  KEY_LEFT   0xff51
#define  KEY_ESCAPE 0xff1b

    case KEY_ENTER: 
    case KEY_ESCAPE:
    case 113:  // 'q'
    case 32:   // space
        // We're outta here.
        if (entry_exit_movement)
            faces->exit_movement_start();

        initiate_destroy = 1;
        // the z_offset translation starts and is carried out by the
        // render function where we exit when it completes
        break;

    case KEY_LEFT:

        faces->goto_previous();

        break;

    case KEY_RIGHT:

        faces->goto_next();

        break;

    case KEY_DOWN:

        faces->goto_previous();

        break;

    case KEY_UP:

        faces->goto_next();

        break;

    default:
        msgout (DEBUG, "KP: No action for %ld / 0x%lx.\n", key, key);
        break;
    }
} // END key_pressed




static void
load_digit_textures()
{
    char filename[100];
    int i;
    unsigned char *data;
    unsigned int sizex, sizey;

    
    // load digits for screens we don't have yet

    for (i = 0; i < MAX_FACES; i++) {
        
        sprintf (filename, "%s/digits/%u.bmp", base_dir, i + 1);

        // allocate space for texture
        data = new unsigned char [128 * 128 * 3];
        if (data == NULL) {
            msgout(ERROR, "Error allocating space for image");
            end_program(0);
        }
        
        if (!load_image(filename, data, &sizex, &sizey)) {
            //msgout (DEBUG, "no image for %u\n", i);
            digit_text[i] = -1;
            continue;
        }
        
        // Create Texture
        glGenTextures(1, (GLuint *)&digit_text[i]);
        glBindTexture(GL_TEXTURE_2D, digit_text[i]);   // 2d texture (x and y size)
        
        glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, MIN_FILTER);
        glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, MAX_FILTER);
        
        // 2d texture, 3 components (red, green, blue), x size from image, y size from image, 
        // rgb color data, unsigned byte data, and finally the data itself.
        gluBuild2DMipmaps(GL_TEXTURE_2D, 3, sizex, sizey, GL_RGB, GL_UNSIGNED_BYTE, data);
        //glTexImage2D(GL_TEXTURE_2D, 0, 3, sizex, sizey, 0, GL_RGB, GL_UNSIGNED_BYTE, data);

        delete data;
    }

} // END load_digit_textures


static void
setup_faces (float z_offset_distance, int current_face, int face_count) 
{
    float aspect;
    float face_height, face_width;

    aspect = (float)GLWin.get_screen_width() / (float)GLWin.get_screen_height();
    face_width = 4.0;
    face_height = face_width/aspect;

    last_time = get_milli_time();

    msgout (DEBUG, "Current area = %d\n", current_face);

    switch (face_type) {
    case FACE_TYPE_CYL:
        faces = new CylinderFaceSet     (face_count, current_face, 
                                         face_width, face_height);
                                         //z_offset_distance);
                                         //5.0);
        break;
    case FACE_TYPE_LINEAR:
        faces = new LinearFaceSet       (face_count, current_face, 
                                         face_width, face_height);
                                         //z_offset_distance);
                                         //10.0);
        break;
    case FACE_TYPE_VIEWMASTER:
        faces = new ViewMasterFaceSet   (face_count, current_face, 
                                         face_width, face_height);
                                         //z_offset_distance);
                                         //10.0);
        break;
    case FACE_TYPE_PRICEISRIGHT:
        faces = new PriceIsRightFaceSet (face_count, current_face, 
                                         face_width, face_height);
                                         //z_offset_distance);
                                         //7.0);
        break;
    case FACE_TYPE_CAROUSEL:
        faces = new CarouselFaceSet     (face_count, current_face, 
                                         face_width, face_height);
                                         //z_offset_distance);
                                         //7.0);
        break;
    default:
        msgout (ERROR, "Unknown face type (?)\n");
        end_program(1);
    }

    if (!faces) {
        msgout(ERROR, "out of memory\n");
        end_program (1);
    }

    if (z_offset_distance)
        faces->set_z_offset_distance(z_offset_distance);

} // END setup_faces

static void 
idle_func(void)
{
    // This prevents the scene from being rendered more then 1000
    // times a second ie 1000fps -- by limiting it like this we sneak
    // in some context switches so the CPU usage isn't through the
    // roof.

    long start = get_milli_time();

    if ( (start - last_renderscene_time) >= 1 ) {

        //msgout (DEBUG, "diff = %lu\n", start - last_renderscene_time);

        render_scene();

    } else {
        //sleep_ms(1);  // allow context switch
        context_switch();
    }


} // END idle_func


// ------
// Does everything needed before losing control to the main
// OpenGL event loop.  

static void
gl_init (void)
{
    // Color to clear color buffer to.
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

    // Depth to clear depth buffer to; type of test.
    glClearDepth(1.0);
    glDepthFunc(GL_LESS); 

    // Enables Smooth Color Shading; try GL_FLAT for (lack of) fun.
    glShadeModel(GL_SMOOTH);

    glLoadIdentity();

    // A handy trick -- have surface material mirror the color.
    //glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);
    //glEnable(GL_COLOR_MATERIAL);

    glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);

    //glLightModelfv(GL_LIGHT_MODEL_AMBIENT, global_ambient);

    if (use_breathing)
        breathing.init(&ambient_value, 1.7, 3.0, 200, Movement::MOVE_TYPE_SINE);
    else
        light_ambient2[0] = light_ambient2[1] = light_ambient2[2] = 2.0;


} // END gl_init

