//  ----------------------------------------------------------
//
//  Copyright (C) 2002 Brad Wasson <bard@systemtoolbox.com>
//
//  This file is part of 3ddesktop.
//
//  3ddesktop is free software; you can redistribute it and/or modify it
//  under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2, or (at your option)
//  any later version.
//
//  3ddesktop is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with 3ddesktop; see the file COPYING.   If not, write to
//  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//

#ifndef _FACE_HPP
#define _FACE_HPP

#include <GL/gl.h>
#include "util.h"

class Point {
public:
    vertex_t v;
    Point (float _x = 0, float _y = 0, float _z = 0) {
        v[0] = _x; v[1] = _y; v[2] = _z; v[3] = 1;
    }
    void set (float _x, float _y, float _z) {
        v[0] = _x; v[1] = _y; v[2] = _z; v[3] = 1;
    }
};


class Face {
public:
    Point topleft, topright, bottomleft, bottomright;
    
    Point normal;

    Face (Point tl, 
          Point tr, 
          Point bl, 
          Point br) {
        topleft = tl;
        topright = tr;
        bottomleft = bl;
        bottomright = br;

        // tl ---- tr
        //         |
        //         br

        float x1 = tl.v[0] - tr.v[0];
        float y1 = tl.v[1] - tr.v[1];
        float z1 = tl.v[2] - tr.v[2];

        float x2 = br.v[0] - tr.v[0];
        float y2 = br.v[1] - tr.v[1];
        float z2 = br.v[2] - tr.v[2];

        normal.v[0] = (y1 * z2) - (y2 * z1);
	normal.v[1] = (z1 * x2) - (z2 * x1);
	normal.v[2] = (x1 * y2) - (x2 * y1);

        float len = (float)sqrt(normal.v[0] * normal.v[0] + 
                                normal.v[1] * normal.v[1] + 
                                normal.v[2] * normal.v[2]);
        normal.v[0] /= len;
        normal.v[1] /= len;
        normal.v[2] /= len;
    }

    void render (int use_texture) {
        glBegin(GL_QUADS);

        //printf ("0x%x\n", &bottomleft);

        glNormal3fv( normal.v );

        if (use_texture)
            glTexCoord2f (0.0f, 0.0f);
        glVertex3fv(topleft.v);

        if (use_texture)
            glTexCoord2f (1.0f, 0.0f);
        glVertex3fv(topright.v);

        if (use_texture)
            glTexCoord2f (1.0f, 1.0f);
        glVertex3fv(bottomright.v);

        if (use_texture)
            glTexCoord2f (0.0f, 1.0f);
        glVertex3fv(bottomleft.v);

        glEnd();

    }

};


#if 0
        if (use_texture)
            glTexCoord2f (0.0f, 0.0f);

        //glVertex3f(bottomleft.v[0],  bottomleft.v[1],  bottomleft.v[2]);
        glVertex3fv(bottomleft.v);

        if (use_texture)
            glTexCoord2f (1.0f, 0.0f);
        //glVertex3f(bottomright.v[0], bottomright.v[1], bottomright.v[2]);
        glVertex3fv(bottomright.v);

        if (use_texture)
            glTexCoord2f (1.0f, 1.0f);
        //glVertex3f(topright.v[0],    topright.v[1],    topright.v[2]);
        glVertex3fv(topright.v);

        if (use_texture)
            glTexCoord2f (0.0f, 1.0f);
        //glVertex3f(topleft.v[0],     topleft.v[1],     topleft.v[2]);
        glVertex3fv(topleft.v);

#endif


#endif // _FACE_HPP
